package sun.nio.fs;

import java.nio.file.StandardWatchEventKinds;
import java.nio.file.NotDirectoryException;
import com.sun.nio.file.SensitivityWatchEventModifier;
import com.sun.nio.file.ExtendedWatchEventModifier;
import java.util.HashMap;
import java.util.Map;
import sun.misc.Unsafe;
import java.util.Set;
import java.nio.file.WatchKey;
import java.nio.file.WatchEvent;
import java.nio.file.Path;
import java.io.IOException;

class WindowsWatchService extends AbstractWatchService
{
    private static final int WAKEUP_COMPLETION_KEY = 0;
    private final WindowsWatchService.Poller poller;
    private static final int ALL_FILE_NOTIFY_EVENTS = 351;
    
    WindowsWatchService(WindowsFileSystem windowsFileSystem) throws IOException {
        long createIoCompletionPort;
        try {
            createIoCompletionPort = WindowsNativeDispatcher.CreateIoCompletionPort(-1L, 0L, 0L);
        }
        catch (WindowsException ex) {
            throw new IOException(ex.getMessage());
        }
        (this.poller = new WindowsWatchService.Poller(windowsFileSystem, this, createIoCompletionPort)).start();
    }
    
    @Override
    WatchKey register(Path path, WatchEvent.Kind<?>[] array, WatchEvent.Modifier... array2) throws IOException {
        return this.poller.register(path, array, array2);
    }
    
    @Override
    void implClose() throws IOException {
        this.poller.close();
    }
    
    private static class WindowsWatchKey extends AbstractWatchKey
    {
        private final WindowsWatchService.FileKey fileKey;
        private volatile long handle;
        private Set<? extends WatchEvent.Kind<?>> events;
        private boolean watchSubtree;
        private NativeBuffer buffer;
        private long countAddress;
        private long overlappedAddress;
        private int completionKey;
        private boolean errorStartingOverlapped;
        
        WindowsWatchKey(Path path, AbstractWatchService abstractWatchService, WindowsWatchService.FileKey fileKey) {
            super(path, abstractWatchService);
            this.handle = -1L;
            this.fileKey = fileKey;
        }
        
        WindowsWatchService.WindowsWatchKey init(long handle, Set<? extends WatchEvent.Kind<?>> events, boolean watchSubtree, NativeBuffer buffer, long countAddress, long overlappedAddress, int completionKey) {
            this.handle = handle;
            this.events = events;
            this.watchSubtree = watchSubtree;
            this.buffer = buffer;
            this.countAddress = countAddress;
            this.overlappedAddress = overlappedAddress;
            this.completionKey = completionKey;
            return this;
        }
        
        long handle() {
            return this.handle;
        }
        
        Set<? extends WatchEvent.Kind<?>> events() {
            return this.events;
        }
        
        void setEvents(Set<? extends WatchEvent.Kind<?>> events) {
            this.events = events;
        }
        
        boolean watchSubtree() {
            return this.watchSubtree;
        }
        
        NativeBuffer buffer() {
            return this.buffer;
        }
        
        long countAddress() {
            return this.countAddress;
        }
        
        long overlappedAddress() {
            return this.overlappedAddress;
        }
        
        WindowsWatchService.FileKey fileKey() {
            return this.fileKey;
        }
        
        int completionKey() {
            return this.completionKey;
        }
        
        void setErrorStartingOverlapped(boolean errorStartingOverlapped) {
            this.errorStartingOverlapped = errorStartingOverlapped;
        }
        
        boolean isErrorStartingOverlapped() {
            return this.errorStartingOverlapped;
        }
        
        void invalidate() {
            WindowsWatchService.Poller.access$100(((WindowsWatchService)this.watcher()).poller, this);
            this.handle = -1L;
            this.buffer = null;
            this.countAddress = 0L;
            this.overlappedAddress = 0L;
            this.errorStartingOverlapped = false;
        }
        
        @Override
        public boolean isValid() {
            return this.handle != -1L;
        }
        
        @Override
        public void cancel() {
            if (this.isValid()) {
                ((WindowsWatchService)this.watcher()).poller.cancel(this);
            }
        }
    }
    
    private static class FileKey
    {
        private final int volSerialNumber;
        private final int fileIndexHigh;
        private final int fileIndexLow;
        
        FileKey(int volSerialNumber, int fileIndexHigh, int fileIndexLow) {
            this.volSerialNumber = volSerialNumber;
            this.fileIndexHigh = fileIndexHigh;
            this.fileIndexLow = fileIndexLow;
        }
        
        @Override
        public int hashCode() {
            return this.volSerialNumber ^ this.fileIndexHigh ^ this.fileIndexLow;
        }
        
        @Override
        public boolean equals(Object o) {
            if (o == this) {
                return true;
            }
            if (!(o instanceof WindowsWatchService.FileKey)) {
                return false;
            }
            WindowsWatchService.FileKey fileKey = (WindowsWatchService.FileKey)o;
            return this.volSerialNumber == fileKey.volSerialNumber && this.fileIndexHigh == fileKey.fileIndexHigh && this.fileIndexLow == fileKey.fileIndexLow;
        }
    }
    
    private static class Poller extends AbstractPoller
    {
        private static final Unsafe UNSAFE;
        private static final short SIZEOF_DWORD = 4;
        private static final short SIZEOF_OVERLAPPED = 32;
        private static final short OFFSETOF_HEVENT;
        private static final short OFFSETOF_NEXTENTRYOFFSET = 0;
        private static final short OFFSETOF_ACTION = 4;
        private static final short OFFSETOF_FILENAMELENGTH = 8;
        private static final short OFFSETOF_FILENAME = 12;
        private static final int CHANGES_BUFFER_SIZE = 16384;
        private final WindowsFileSystem fs;
        private final WindowsWatchService watcher;
        private final long port;
        private final Map<Integer, WindowsWatchService.WindowsWatchKey> ck2key;
        private final Map<WindowsWatchService.FileKey, WindowsWatchService.WindowsWatchKey> fk2key;
        private int lastCompletionKey;
        
        Poller(WindowsFileSystem fs, WindowsWatchService watcher, long port) {
            this.fs = fs;
            this.watcher = watcher;
            this.port = port;
            this.ck2key = new HashMap();
            this.fk2key = new HashMap();
            this.lastCompletionKey = 0;
        }
        
        @Override
        void wakeup() throws IOException {
            try {
                WindowsNativeDispatcher.PostQueuedCompletionStatus(this.port, 0L);
            }
            catch (WindowsException ex) {
                throw new IOException(ex.getMessage());
            }
        }
        
        @Override
        Object implRegister(Path path, Set<? extends WatchEvent.Kind<?>> events, WatchEvent.Modifier... array) {
            WindowsPath windowsPath = (WindowsPath)path;
            boolean n = false;
            for (WatchEvent.Modifier modifier : array) {
                if (modifier == ExtendedWatchEventModifier.FILE_TREE) {
                    n = true;
                }
                else {
                    if (modifier == null) {
                        return new NullPointerException();
                    }
                    if (!(modifier instanceof SensitivityWatchEventModifier)) {
                        return new UnsupportedOperationException("Modifier not supported");
                    }
                }
            }
            long createFile;
            try {
                createFile = WindowsNativeDispatcher.CreateFile(windowsPath.getPathForWin32Calls(), 1, 7, 3, 1107296256);
            }
            catch (WindowsException ex) {
                return ex.asIOException(windowsPath);
            }
            int n2 = 0;
            try {
                WindowsFileAttributes attributes;
                try {
                    attributes = WindowsFileAttributes.readAttributes(createFile);
                }
                catch (WindowsException ex2) {
                    IOException ioException = ex2.asIOException(windowsPath);
                }
                if (!attributes.isDirectory()) {
                    NotDirectoryException o4 = new NotDirectoryException(windowsPath.getPathForExceptionMessage());
                }
                else {
                    WindowsWatchService.FileKey o = new WindowsWatchService.FileKey(attributes.volSerialNumber(), attributes.fileIndexHigh(), attributes.fileIndexLow());
                    WindowsWatchService.WindowsWatchKey windowsWatchKey = (WindowsWatchService.WindowsWatchKey)this.fk2key.get(o);
                    if (windowsWatchKey != null && n == (windowsWatchKey.watchSubtree() ? 1 : 0)) {
                        windowsWatchKey.setEvents(events);
                    }
                    else {
                        int j = ++this.lastCompletionKey;
                        if (j == 0) {
                            j = ++this.lastCompletionKey;
                        }
                        try {
                            WindowsNativeDispatcher.CreateIoCompletionPort(createFile, this.port, (long)j);
                        }
                        catch (WindowsException ex3) {
                            IOException o2 = new IOException(ex3.getMessage());
                        }
                        int n3 = 16420;
                        NativeBuffer nativeBuffer = NativeBuffers.getNativeBuffer(n3);
                        long address = nativeBuffer.address();
                        long address2 = address + (long)n3 - 32L;
                        long n4 = address2 - 4L;
                        WindowsWatchService.Poller.UNSAFE.setMemory(address2, 32L, (byte)0);
                        try {
                            this.createAndAttachEvent(address2);
                            WindowsNativeDispatcher.ReadDirectoryChangesW(createFile, address, 16384, (boolean)n, 351, n4, address2);
                        }
                        catch (WindowsException ex4) {
                            this.closeAttachedEvent(address2);
                            nativeBuffer.release();
                            IOException o3 = new IOException(ex4.getMessage());
                        }
                        WindowsWatchService.WindowsWatchKey windowsWatchKey2;
                        if (windowsWatchKey == null) {
                            windowsWatchKey2 = new WindowsWatchService.WindowsWatchKey(windowsPath, this.watcher, o).init(createFile, events, (boolean)n, nativeBuffer, n4, address2, j);
                            this.fk2key.put(o, windowsWatchKey2);
                        }
                        else {
                            this.ck2key.remove(Integer.valueOf(windowsWatchKey.completionKey()));
                            this.releaseResources(windowsWatchKey);
                            windowsWatchKey2 = windowsWatchKey.init(createFile, events, (boolean)n, nativeBuffer, n4, address2, j);
                        }
                        this.ck2key.put(Integer.valueOf(j), windowsWatchKey2);
                        n2 = 1;
                    }
                }
            }
            finally {
                if (n2 == 0) {
                    WindowsNativeDispatcher.CloseHandle(createFile);
                }
            }
        }
        
        private void releaseResources(WindowsWatchService.WindowsWatchKey windowsWatchKey) {
            if (!windowsWatchKey.isErrorStartingOverlapped()) {
                try {
                    WindowsNativeDispatcher.CancelIo(windowsWatchKey.handle());
                    WindowsNativeDispatcher.GetOverlappedResult(windowsWatchKey.handle(), windowsWatchKey.overlappedAddress());
                }
                catch (WindowsException ex) {}
            }
            WindowsNativeDispatcher.CloseHandle(windowsWatchKey.handle());
            this.closeAttachedEvent(windowsWatchKey.overlappedAddress());
            windowsWatchKey.buffer().cleaner().clean();
        }
        
        private void createAndAttachEvent(long n) throws WindowsException {
            WindowsWatchService.Poller.UNSAFE.putAddress(n + (long)WindowsWatchService.Poller.OFFSETOF_HEVENT, WindowsNativeDispatcher.CreateEvent(false, false));
        }
        
        private void closeAttachedEvent(long n) {
            long address = WindowsWatchService.Poller.UNSAFE.getAddress(n + (long)WindowsWatchService.Poller.OFFSETOF_HEVENT);
            if (address != 0L && address != -1L) {
                WindowsNativeDispatcher.CloseHandle(address);
            }
        }
        
        @Override
        void implCancelKey(WatchKey watchKey) {
            WindowsWatchService.WindowsWatchKey windowsWatchKey = (WindowsWatchService.WindowsWatchKey)watchKey;
            if (windowsWatchKey.isValid()) {
                this.fk2key.remove(windowsWatchKey.fileKey());
                this.ck2key.remove(Integer.valueOf(windowsWatchKey.completionKey()));
                windowsWatchKey.invalidate();
            }
        }
        
        @Override
        void implCloseAll() {
            this.ck2key.values().forEach(WindowsWatchService.WindowsWatchKey::invalidate);
            this.fk2key.clear();
            this.ck2key.clear();
            WindowsNativeDispatcher.CloseHandle(this.port);
        }
        
        private WatchEvent.Kind<?> translateActionToEvent(int n) {
            switch (n) {
                case 3: {
                    return StandardWatchEventKinds.ENTRY_MODIFY;
                }
                case 1:
                case 5: {
                    return StandardWatchEventKinds.ENTRY_CREATE;
                }
                case 2:
                case 4: {
                    return StandardWatchEventKinds.ENTRY_DELETE;
                }
                default: {
                    return null;
                }
            }
        }
        
        private void processEvents(WindowsWatchService.WindowsWatchKey windowsWatchKey, int n) {
            long address = windowsWatchKey.buffer().address();
            int i;
            do {
                WatchEvent.Kind<Object> translateActionToEvent = this.translateActionToEvent(WindowsWatchService.Poller.UNSAFE.getInt(address + 4L));
                if (windowsWatchKey.events().contains(translateActionToEvent)) {
                    int int1 = WindowsWatchService.Poller.UNSAFE.getInt(address + 8L);
                    if (int1 % 2 != 0) {
                        throw new AssertionError("FileNameLength is not a multiple of 2");
                    }
                    char[] array = new char[int1 / 2];
                    WindowsWatchService.Poller.UNSAFE.copyMemory(null, address + 12L, array, (long)Unsafe.ARRAY_CHAR_BASE_OFFSET, (long)int1);
                    windowsWatchKey.signalEvent(translateActionToEvent, WindowsPath.createFromNormalizedPath(this.fs, new String(array)));
                }
                i = WindowsWatchService.Poller.UNSAFE.getInt(address + 0L);
                address += (long)i;
            } while (i != 0);
        }
        
        @Override
        public void run() {
            while (true) {
                WindowsNativeDispatcher.CompletionStatus getQueuedCompletionStatus;
                try {
                    getQueuedCompletionStatus = WindowsNativeDispatcher.GetQueuedCompletionStatus(this.port);
                }
                catch (WindowsException ex) {
                    ex.printStackTrace();
                    return;
                }
                if (getQueuedCompletionStatus.completionKey() == 0L) {
                    if (this.processRequests()) {
                        break;
                    }
                    continue;
                }
                else {
                    WindowsWatchService.WindowsWatchKey windowsWatchKey = (WindowsWatchService.WindowsWatchKey)this.ck2key.get(Integer.valueOf((int)getQueuedCompletionStatus.completionKey()));
                    if (windowsWatchKey == null) {
                        continue;
                    }
                    boolean n = false;
                    int error = getQueuedCompletionStatus.error();
                    int bytesTransferred = getQueuedCompletionStatus.bytesTransferred();
                    if (error == 1022) {
                        windowsWatchKey.signalEvent(StandardWatchEventKinds.OVERFLOW, null);
                    }
                    else if (error != 0 && error != 234) {
                        n = true;
                    }
                    else {
                        if (bytesTransferred > 0) {
                            this.processEvents(windowsWatchKey, bytesTransferred);
                        }
                        else if (error == 0) {
                            windowsWatchKey.signalEvent(StandardWatchEventKinds.OVERFLOW, null);
                        }
                        try {
                            WindowsNativeDispatcher.ReadDirectoryChangesW(windowsWatchKey.handle(), windowsWatchKey.buffer().address(), 16384, windowsWatchKey.watchSubtree(), 351, windowsWatchKey.countAddress(), windowsWatchKey.overlappedAddress());
                        }
                        catch (WindowsException ex2) {
                            n = true;
                            windowsWatchKey.setErrorStartingOverlapped(true);
                        }
                    }
                    if (!(n == 0)) {
                        this.implCancelKey(windowsWatchKey);
                        windowsWatchKey.signal();
                    }
                    else {
                        continue;
                    }
                }
            }
        }
        
        static void access$100(WindowsWatchService.Poller poller, WindowsWatchService.WindowsWatchKey windowsWatchKey) {
            poller.releaseResources(windowsWatchKey);
        }
        
        static {
            UNSAFE = Unsafe.getUnsafe();
            OFFSETOF_HEVENT = ((WindowsWatchService.Poller.UNSAFE.addressSize() == 4) ? 16 : 24);
        }
    }
}
